
/*  $Id: HBCIJob.java,v 1.1 2011/05/04 22:37:53 willuhn Exp $

    This file is part of HBCI4Java
    Copyright (C) 2001-2008  Stefan Palme

    HBCI4Java is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    HBCI4Java is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package org.kapott.hbci.GV;

import java.util.Date;
import java.util.List;
import java.util.Properties;

import org.kapott.hbci.GV_Result.HBCIJobResult;
import org.kapott.hbci.passport.HBCIPassport;
import org.kapott.hbci.structures.Konto;
import org.kapott.hbci.structures.Value;

/** <p>Schnittstelle fr alle Auftrge, die via HBCI ausgefhrt werden sollen. Ein
 HBCIJob-Objekt wird nur innerhalb von <em>HBCI4Java</em> verwaltet. Durch Aufruf einer der Methoden
 {@link org.kapott.hbci.manager.HBCIHandler#newJob(String)} oder
 {@link org.kapott.hbci.manager.HBCIHandler#newLowlevelJob(String)} wird
 eine neue Instanz eines HBCIJobs erzeugt. Die konkrete Klasse dieser Instanz ist
 fr den Anwendungsentwickler nicht von Bedeutung.</p>
 <p>Die Anwendung muss nur die fr diesen Job bentigten Parameter setzen (mit 
 {@link #setParam(String,String)}). Falls dieser Job mehrere digitale
 Signaturen bentigt, knnen mit der Methode {@link #addSignaturePassport(HBCIPassport,String)} 
 weitere Passport-Objekte zu diesem Job hinzugefgt werden,
 die dann als Zweit-, Dritt-, ...-Signatur bei der Nachrichtenerzeugung verwendet
 werden. Anschlieend kann der fertig spezifizierte Job zum aktuellen HBCI-Dialog 
 hinzugefgt werden 
 ({@link #addToQueue()}).</p>
 <p>Nach Ausfhrung des HBCI-Dialoges knnen die Rckgabedaten und Statusinformationen fr diesen
 Job ermittelt werden. Dazu wird die Methoode {@link #getJobResult()} bentigt, welche
 eine Instanz einer {@link org.kapott.hbci.GV_Result.HBCIJobResult}-Klasse zurckgibt.
 Die konkrete Klasse, um die es sich bei diesem Result-Objekt handelt, ist vom Typ des ausgefhrten
 Jobs abhngig (z.B. gibt es eine Klasse, die Ergebnisdaten fr Kontoauszge enthlt, eine
 Klasse fr Saldenabfragen usw.). Eine Beschreibung der einzelnen Klassen fr Result-Objekte findet
 sich im Package <code>org.kapott.hbci.GV_Result</code>. Eine Beschreibung, welcher Job welche Klasse
 zurckgibt, befindet sich in der Package-Dokumentation zu diesem Package (<code>org.kapott.hbci.GV</code>).</p> */
public interface HBCIJob
{
    /** Gibt den internen Namen fr diesen Job zurck.
     * @return Job-Name, wie er intern von <em>HBCI4Java</em> verwendet wird. */
    public String getName();
    
    /** Gibt die fr diesen Job verwendete Segment-Versionsnummer zurck */
    public String getSegVersion();

    /** <p>Gibt zurck, wieviele Signaturen fr diesen Job mindestens bentigt werden.
     *  Diese Information wird den BPD entnommen. In einigen Fllen gibt es
     *  in den UPD aktuellere Informationen zu einem bestimmten Geschftsvorfall,
     *  zur Zeit werden die UPD von dieser Methode aber nicht ausgewertet.</p>
     *  <p>Wird fr einen Job mehr als eine Signatur bentigt, so knnen mit der
     *  Methode {@link #addSignaturePassport(HBCIPassport, String)}
     *  Passports bestimmt werden, die fr die Erzeugung der zustzlichen
     *  Signaturen verwendet werden sollen.</p>
     *  <p>Es wird auerdem empfohlen, dass Auftrge, die mehrere Signaturen
     *  bentigen, jeweils in einer separaten HBCI-Nachricht versandt werden. Um das
     *  zu erzwingen, kann entweder ein HBCI-Dialog gefhrt werden, der definitiv
     *  nur diesen einen Auftrag enthlt (also nur ein 
     *  {@link #addToQueue()}
     *  fr diesen Dialog), oder es wird beim Zusammenstellen der Jobs fr einen
     *  Dialog sichergestellt, dass ein bestimmter Job in einer separaten Nachricht
     *  gesandt wird 
     *  ({@link org.kapott.hbci.manager.HBCIHandler#newMsg()}).</p>
     *  @return Mindest-Anzahl der bentigten Signaturen fr diesen Job */
    public int getMinSigs();

    /** <p>Gibt zurck, welche Sicherheitsklasse fr diesen Job mindestens bentigt wird.
     *  Diese Information wird den BPD entnommen. Sicherheitsklassen sind erst
     *  ab FinTS-3.0 definiert. Falls keine Sicherheitsklassen untersttzt werden
     *  (weil eine geringere HBCI-Version als FinTS-3.0 verwendet wird), wird 
     *  <code>1</code> zurckgegeben. Die Sicherheitsklasse ist nur die 
     *  Sicherheitsmechanismen DDV und RDH relevant - bei Verwendung von PIN/TAN
     *  hat die Sicherheitsklasse keine Bedeutung. </p>
     *  <p>Folgende Sicherheitsklassen sind definiert:</p>
     *  <ul>
     *  <li><code>0</code>: kein Sicherheitsdienst erforderlich</li>
     *  <li><code>1</code>: Authentication - es wird eine Signatur mit dem
     *  Signaturschlssel bentigt.</li>
     *  <li><code>2</code>: Authentication mit fortgeschrittener elektronischer
     *  Signatur unter Verwendung des Signaturschlssels.</li>
     *  <li><code>3</code>: Non-Repudiation mit fortgeschrittener elektronischer
     *  Signatur und optionaler Zertifikatsprfung unter Verwendung des DigiSig-Schlssels</li>
     *  <li><code>4</code>: Non-Repudiation mit fortgeschrittener bzw. qualifizierter
     *  elektronischer Signatur und zwingender Zertifikatsberprfung mit dem
     *  DigiSig-Schlssel</li>
     *  </ul>
     *  @return Sicherheitsklasse, die fr diese Job bentigt wird */
    public int getSecurityClass();

    /** Gibt alle mglichen Job-Parameter fr einen Lowlevel-Job zurck.
     * Die Anwendung dieser Methode ist nur sinnvoll, wenn es sich bei dem 
     * aktuellen Job um einen Lowlevel-Job handelt (erzeugt mit 
     * {@link org.kapott.hbci.manager.HBCIHandler#newLowlevelJob(String)}).
     * Die zurckgegebenen Parameternamen knnen als erstes Argument der
     * Methode {@link #setParam(String, String)} verwendet werden. 
     * @return Liste aller gltigen Parameternamen (nur fr Lowlevel-Jobs) */
    public List getJobParameterNames();

    /** Gibt alle mglichen Property-Namen fr die Lowlevel-Rckgabedaten dieses
     * Jobs zurck. Die Lowlevel-Rckgabedaten knnen mit
     * {@link #getJobResult()} und {@link HBCIJobResult#getResultData()}
     * ermittelt werden. Diese Methode verwendet intern
     * {@link org.kapott.hbci.manager.HBCIHandler#getLowlevelJobResultNames(String)}.
     * @return Liste aller prinzipiell mglichen Property-Keys fr die 
     * Lowlevel-Rckgabedaten dieses Jobs */
    public List getJobResultNames();

    /** <p>Gibt fr einen Job alle bekannten Einschrnkungen zurck, die bei
     der Ausfhrung des jeweiligen Jobs zu beachten sind. Diese Daten werden aus den
     Bankparameterdaten des aktuellen Passports extrahiert. Sie knnen von einer HBCI-Anwendung
     benutzt werden, um gleich entsprechende Restriktionen bei der Eingabe von
     Geschftsvorfalldaten zu erzwingen (z.B. die maximale Anzahl von Verwendungszweckzeilen,
     ob das ndern von terminierten berweisungen erlaubt ist usw.).</p>
     <p>Die einzelnen Eintrge des zurckgegebenen Properties-Objektes enthalten als Key die
     Bezeichnung einer Restriktion (z.B. "<code>maxusage</code>"), als Value wird der
     entsprechende Wert eingestellt. Die Bedeutung der einzelnen Restriktionen ist zur Zeit
     nur der HBCI-Spezifikation zu entnehmen. In spteren Programmversionen werden entsprechende
     Dokumentationen zur internen HBCI-Beschreibung hinzugefgt, so dass dafr eine Abfrageschnittstelle
     implementiert werden kann.</p>
     <p>Diese Methode verwendet intern 
     {@link org.kapott.hbci.manager.HBCIHandler#getLowlevelJobRestrictions(String)}</p>.
     @return Properties-Objekt mit den einzelnen Restriktionen */
    public Properties getJobRestrictions();

    /** Gibt alle fr diesen Job gesetzten Parameter zurck. In dem
     zurckgegebenen <code>Properties</code>-Objekt sind werden die
     Parameter als <em>Lowlevel</em>-Parameter abgelegt. Auerdem hat
     jeder Lowlevel-Parametername zustzlich ein Prefix, welches den
     Lowlevel-Job angibt, fr den der Parameter gilt (also z.B.
     <code>Ueb3.BTG.value</code>
     @return aktuelle gesetzte Lowlevel-Parameter fr diesen Job */
    public Properties getLowlevelParams();

    /** Setzen eines komplexen Job-Parameters (Kontodaten). Einige Jobs bentigten Kontodaten
     als Parameter. Diese mssten auf "normalem" Wege durch mehrere Aufrufe von 
     {@link #setParam(String,String)} erzeugt werden (Lnderkennung, Bankleitzahl, 
     Kontonummer, Unterkontomerkmal, Whrung, IBAN, BIC).
     Durch Verwendung dieser Methode wird dieser Weg abgekrzt. Es wird ein Kontoobjekt 
     bergeben, fr welches die entsprechenden <code>setParam(String,String)</code>-Aufrufe 
     automatisch erzeugt werden. 
     @param paramname die Basis der Parameter fr die Kontodaten (fr <code>my.country</code>,
     <code>my.blz</code>, <code>my.number</code>, <code>my.subnumber</code>, <code>my.bic</code>, 
     <code>my.iban</code>, <code>my.curr</code> wre das also "<code>my</code>")
     @param acc ein Konto-Objekt, aus welchem die zu setzenden Parameterdaten entnommen werden */
    public void setParam(String paramname,Konto acc);

    /** Setzen eines komplexen Job-Parameters (Geldbetrag). Einige Jobs bentigten Geldbetrge
     als Parameter. Diese mssten auf "normalem" Wege durch zwei Aufrufe von 
     {@link #setParam(String,String)} erzeugt werden (je einer fr
     den Wert und die Whrung). Durch Verwendung dieser
     Methode wird dieser Weg abgekrzt. Es wird ein Value-Objekt bergeben, fr welches
     die entsprechenden zwei <code>setParam(String,String)</code>-Aufrufe automatisch
     erzeugt werden.
     @param paramname die Basis der Parameter fr die Geldbetragsdaten (fr "<code>btg.value</code>" und
     "<code>btg.curr</code>" wre das also "<code>btg</code>")
     @param v ein Value-Objekt, aus welchem die zu setzenden Parameterdaten entnommen werden */
    public void setParam(String paramname,Value v);

    /** Setzen eines Job-Parameters, bei dem ein Datums als Wert erwartet wird. Diese Methode
     dient als Wrapper fr {@link #setParam(String,String)}, um das Datum in einen korrekt
     formatierten String umzuwandeln. Das "richtige" Datumsformat ist dabei abhngig vom
     aktuellen Locale.
     @param paramName Name des zu setzenden Job-Parameters
     @param date Datum, welches als Wert fr den Job-Parameter benutzt werden soll */
    public void setParam(String paramName,Date date);

    /** Setzen eines Job-Parameters, bei dem ein Integer-Wert Da als Wert erwartet wird. Diese Methode
     dient nur als Wrapper fr {@link #setParam(String,String)}.
     @param paramName Name des zu setzenden Job-Parameters
     @param i Integer-Wert, der als Wert gesetzt werden soll */
    public void setParam(String paramName,int i);

    /** <p>Setzen eines Job-Parameters. Fr alle Highlevel-Jobs ist in der Package-Beschreibung zum
     Package <code>org.kapott.hbci.GV</code> eine Auflistung aller Jobs und deren Parameter zu finden.
     Fr alle Lowlevel-Jobs kann eine Liste aller Parameter entweder mit dem Tool
     {@link org.kapott.hbci.tools.ShowLowlevelGVs} oder zur Laufzeit durch Aufruf
     der Methode {@link org.kapott.hbci.manager.HBCIHandler#getLowlevelJobParameterNames(String)} 
     ermittelt werden.</p>
     <p>Bei Verwendung dieser oder einer der anderen <code>setParam()</code>-Methoden werden zustzlich
     einige der Job-Restriktionen (siehe {@link #getJobRestrictions()}) analysiert. Beim Verletzen einer
     der berprften Einschrnkungen wird eine Exception mit einer entsprechenden Meldung erzeugt.
     Diese berprfung findet allerdings nur bei Highlevel-Jobs statt.</p>
     @param paramName der Name des zu setzenden Parameters.
     @param value Wert, auf den der Parameter gesetzt werden soll */
    public void setParam(String paramName,String value);
    
    /** <p>Hinzufgen dieses Jobs zu einem HBCI-Dialog. Diese Methode arbeitet analog zu 
        {@link #addToQueue(String)}, nur dass hier
        die <code>customerid</code> mit der Kunden-ID vorbelegt ist, wie sie
        im aktuellen Passport gespeichert ist.</p> */
    public void addToQueue();
    
    /** <p>Hinzufgen dieses Jobs zu einem HBCI-Dialog. Nachdem alle
        Jobparameter mit 
        {@link #setParam(String,String)}
        gesetzt wurden, kann der komplett spezifizierte Job mit dieser Methode
        zur Auftragsliste eines Dialoges hinzugefgt werden.</p>
        <p>Die <code>customerId</code> gibt an, unter welcher Kunden-ID dieser Job
        ausgefhrt werden soll. Existiert fr eine HBCI-Nutzerkennung (ein Passport)
        nur genau eine Kunden-ID (wie das i.d.R. der Fall ist), so kann der
        <code>customerId</code>-Parameter weggelassen werden - <em>HBCI4Java</em>
        verwendet dann automatisch die richtige Kunden-ID (als Kunden-ID wird in diesem
        Fall der Wert von {@link HBCIPassport#getCustomerId()} verwendet). Gibt es aber mehrere
        gltige Kunden-IDs fr einen HBCI-Zugang, so muss die Kunden-ID,
        die fr diesen Job verwendet werden soll, mit angegeben werden.</p>
        <p>Jeder Auftrag (=Job) ist i.d.R. an ein bestimmtes Konto des Auftraggebers
        gebunden (berweisung: das Belastungskonto; Saldenabfrage: das abzufragende
        Konto usw.). Als Kunden-ID fr einen Auftrag muss <em>die</em> Kunden-ID
        angegeben werden, die fr dieses Konto verfgungsberechtigt ist.</p>
        <p>I.d.R. liefert eine Bank Informationen ber alle Konten, auf die
        via HBCI zugegriffen werden kann. Ist das der Fall, so kann die Menge
        dieser Konten mit {@link HBCIPassport#getAccounts()} ermittelt werden.
        In jedem zurckgemeldeten {@link org.kapott.hbci.structures.Konto}-Objekt 
        ist im Feld <code>customerid</code> vermerkt, welche Kunden-ID fr 
        dieses Konto verfgungsberechtigt ist. Diese Kunden-ID msste dann also 
        beim Hinzufgen eines Auftrages angegeben werden, welcher das jeweilige 
        Konto betrifft.</p>
        <p>Liefert eine Bank diese Informationen nicht, so hat die Anwendung selbst
        eine Kontenverwaltung zu implementieren, bei der jedem Nutzerkonto eine
        zu verwendende Kunden-ID zugeordnet ist.</p>
        <p>Ein HBCI-Dialog kann aus beliebig vielen HBCI-Nachrichten bestehen. <em>HBCI4Java</em> versucht zunchst,
        alle Jobs in einer einzigen Nachricht unterzubringen. Kann ein Job nicht mehr zur aktuellen
        Nachricht hinzugefgt werden (weil sonst bestimmte vorgegebene Bedingungen nicht eingehalten
        werden), so legt <em>HBCI4Java</em> automatisch eine neue Nachricht an, zu der der Job schlielich
        hinzugefgt wird. Beim Ausfhren des HBCI-Dialoges (siehe {@link org.kapott.hbci.manager.HBCIHandler#execute()}) werden dann
        natrlich <em>alle</em> erzeugten Nachrichten zum HBCI-Server gesandt.</p> 
        <p>Der HBCI-Kernel bestimmt also automatisch, ob ein Auftrag noch mit in die aktuelle Nachricht
        aufgenommen werden kann, oder ob eine separate Nachricht erzeugt werden muss. Der manuelle
        Aufruf von {@link org.kapott.hbci.manager.HBCIHandler#newMsg() HBCIHandler.newMsg()} ist 
        deshalb im Prinzip niemals notwendig, es sei denn,
        es soll aus anderen Grnden eine neue Nachricht begonnen werden.</p>
        @param customerId die Kunden-ID, zu deren Dialog der Auftrag hinzugefgt werden soll */
    public void addToQueue(String customerId);

    /** Gibt ein Objekt mit den Rckgabedaten fr diesen Job zurck. Das zurckgegebene Objekt enthlt
     erst <em>nach</em> der Ausfhrung des Jobs gltige Daten.
     @return ein Objekt mit den Rckgabedaten und Statusinformationen zu diesem Job */
    public HBCIJobResult getJobResult();

    /** Hinzufgen eines Passports, welches fr eine zustzliche Signatur fr
     *  diesen Auftrag benutzt wird. <code>role</code> gibt dabei die Rolle an,
     *  die der Eigentmer des zustzlichen Passports in Bezug auf diesen
     *  Job (bzw. die aktuelle Nachricht) einnimmt. Gltige Werte sind in
     *  {@link org.kapott.hbci.passport.HBCIPassport} beschrieben.
     *  Mit der Methode {@link #getMinSigs()} kann ermittelt werden, wieviele
     *  Signaturen fr einen Job mindestens bentigt werden. 
     *  @param passport das hinzuzufgende Passport-Objekt, welches fr eine 
     *         zustzliche Signatur benutzt werden soll
     *  @param role die Rolle, in der sich der Eigentmer des zustzlichen 
     *         Passport-Objektes bezglich dieses Jobs befindet */
    public void addSignaturePassport(HBCIPassport passport,String role);
}
