
/*  $Id: HBCIExecStatus.java 62 2008-10-22 17:03:26Z kleiner $

    This file is part of HBCI4Java
    Copyright (C) 2001-2008  Stefan Palme

    HBCI4Java is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    HBCI4Java is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package org.kapott.hbci.status;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.kapott.hbci.manager.HBCIUtils;

/** Statusinformationen ber alle ausgefhrten Dialoge. Die Methode
    {@link org.kapott.hbci.manager.HBCIHandler#execute()} gibt nach der Ausfhrung 
    aller HBCI-Dialoge ein Objekt dieser Klasse zurck. Dieses Objekt enthlt 
    Informationen darber, fr welche Kunden-IDs tatschlich HBCI-Dialoge gefhrt 
    wurden. Fr jeden gefhrten HBCI-Dialog existiert dann ein 
    {@link HBCIDialogStatus}-Objekt, welches Informationen zu dem jeweiligen 
    Dialog enthlt. */
public class HBCIExecStatus 
{
    private Map statusData;
    private Map exceptions;
    
    public HBCIExecStatus()
    {
        statusData=new Hashtable();
        exceptions=new Hashtable();
    }
    
    /** Gibt die Menge aller Kunden-IDs zurck, fr die ein HBCI-Dialog gefhrt wurde.
        @return Liste mit Kunden-IDs */
    public List getCustomerIds()
    {
        Set ret=new HashSet();
        
        Set sset=statusData.keySet();
        if (sset!=null) {
            ret.addAll(sset);
        }
        
        Set eset=exceptions.keySet();
        if (eset!=null) {
            ret.addAll(eset);
        }
        
        return new ArrayList(ret);
    }
    
    /** Wird von der <em>HBCI4Java</em>-Dialog-Engine aufgerufen */
    public void addDialogStatus(String customerid,HBCIDialogStatus status)
    {
        if (status!=null) {
            statusData.put(customerid,status);
        } else {
            statusData.remove(customerid);
        }
    }
    
    /** Wird von der <em>HBCI4Java</em>-Dialog-Engine aufgerufen */
    public void addException(String customerid,Exception e)
    {
        ArrayList exc=(ArrayList)exceptions.get(customerid);
        if (exc==null) {
            exc=new ArrayList();
            exceptions.put(customerid,exc);
        }
        exc.add(e);
        HBCIUtils.log(e);
    }
    
    /** Gibt eine Liste von Status-Informationen fr jeden ausgefhrten HBCI-Dialog
        zurck. Diese Methode ist insofern von eingeschrnkter Bedeutung, weil
        es nicht mglich ist, einem {@link HBCIDialogStatus}-Objekt dieser Liste
        die Kunden-ID zuzuordnen, unter der der jeweilige Dialog gefhrt wurde.
        Dazu mssen die Methoden {@link #getCustomerIds()} und {@link #getDialogStatus(String)}
        verwendet werden.
        @return Menge aller gespeicherten HBCI-Dialog-Status-Informationen 
        @deprecated sinnlos */
    public List getDialogStatusList()
    {
        Collection values=statusData.values();
        return (values!=null)?(new ArrayList(values)):(new ArrayList());
    }
    
    /** {@link HBCIDialogStatus} fr den Dialog einer bestimmten Kunden-ID zurckgeben.
        @param customerid die Kunden-ID, fr deren Dialog das Status-Objekt zurckgegeben werden soll
        @return Status-Objekt fr den ausgewhlten Dialog */
    public HBCIDialogStatus getDialogStatus(String customerid)
    {
        return (HBCIDialogStatus)statusData.get(customerid);
    }
    
    /** Exceptions zurckgeben, die beim Ausfhren eines bestimmten Dialoges aufgetreten sind.
        Dabei werden nur die Exceptions zurckgegeben, die Fehler in der Verwaltung der
        Kunden-IDs/Dialoge betreffen. Alle Exceptions, die whrend der eigentlichen
        Dialogausfhrung evtl. aufgetreten sind, sind im entsprechenden
        {@link HBCIDialogStatus}-Objekt des jeweiligen Dialoges enthalten.
        @param customerid die Kunden-ID, fr deren HBCI-Dialog die evtl. aufgetretenen
        Exceptions ermittelt werden sollen.
        @return Liste mit aufgetretenen Exceptions */
    public List getExceptions(String customerid)
    {
        return (ArrayList)exceptions.get(customerid);
    }
    
    /** Gibt einen String zurck, der alle Fehlermeldungen aller ausgefhrten
        Dialog enthlt.
        @return String mit allen aufgetretenen Fehlermeldungen */
    public String getErrorString()
    {
        StringBuffer ret=new StringBuffer();
        String       linesep=System.getProperty("line.separator");
        int          nofCustomerIds=getCustomerIds().size();
        
        for (Iterator i=getCustomerIds().iterator();i.hasNext();) {
            String customerid=(String)i.next();
            boolean customeridWritten=false;
            
            List exc=getExceptions(customerid);
            if (exc!=null && exc.size()!=0) {
                if (nofCustomerIds>1) {
                    ret.append("Dialog for '").append(customerid).append("':").append(linesep);
                    customeridWritten=true;
                }
                
                // ret.append(HBCIUtilsInternal.getLocMsg("STAT_EXCEPTIONS")).append(":").append(linesep);
                for (Iterator j=exc.iterator();j.hasNext();) {
                    ret.append(HBCIUtils.exception2StringShort((Exception)j.next()));
                    ret.append(linesep);
                }
            }
            
            HBCIDialogStatus status=getDialogStatus(customerid);
            if (status!=null) {
                String errMsg=status.getErrorString();
                if (errMsg.length()!=0) {
                    if (nofCustomerIds>1 && !customeridWritten) {
                        ret.append("Dialog for '").append(customerid).append("':").append(linesep);
                        customeridWritten=true;
                    }
                    ret.append(errMsg+linesep);
                }
            }
        }
        
        return ret.toString().trim();
    }
    
    public String toString(String customerId)
    {
        StringBuffer ret=new StringBuffer();
        String       linesep=System.getProperty("line.separator");
        
        List exc=getExceptions(customerId);
        if (exc!=null) {
            for (Iterator j=exc.iterator();j.hasNext();) {
                ret.append(HBCIUtils.exception2StringShort((Exception)j.next()));
                ret.append(linesep);
            }
        }

        HBCIDialogStatus status=getDialogStatus(customerId);
        if (status!=null) {
            ret.append(status.toString()+linesep);
        }
        
        return ret.toString().trim();
    }
    
    /** Gibt einen String mit allen Status-Informationen ber alle ausgefhrten
        Dialoge zurck.
        @return textuelle Darstellung aller gespeicherten Statusdaten */
    public String toString()
    {
        StringBuffer ret=new StringBuffer();
        String       linesep=System.getProperty("line.separator");
        
        for (Iterator i=getCustomerIds().iterator();i.hasNext();) {
            String customerid=(String)i.next();
            ret.append("Dialog for '");
            ret.append(customerid);
            ret.append("':");
            ret.append(linesep);
            ret.append(toString(customerid));
            ret.append(linesep);
        }
        
        return ret.toString().trim();
    }
    
    public boolean isOK(String customerId)
    {
        boolean          ok=true;
        List             exc=getExceptions(customerId);
        HBCIDialogStatus status=getDialogStatus(customerId);
        
        ok&=(exc==null);
        ok&=(status!=null);
        ok&=status.isOK();
        
        return ok;
    }
    
    /** Gibt zurck, ob alle "geplanten" HBCI-Dialoge ordnungsgem ausgefhrt wurden.
        @return <code>false</code>, wenn wenigstens bei einer Dialog-Ausfhrung
        fr eine Kunden-ID ein Fehler aufgetreten ist; ansonsten <code>true</code>*/
    public boolean isOK()
    {
        boolean ok=true;
        List    customerIds=getCustomerIds();
        
        for (Iterator i=customerIds.iterator();i.hasNext();) {
            String customerId=(String)i.next();
            ok&=isOK(customerId);
        }
        
        return ok;
    }
}
